%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Library functions.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module('Corelib', []).

:- use_module(library(iso_byte_char), [char_code/2] ).
:- use_module(library(freeze),        [freeze/2]).
%% This header is an example of how to write a new prolog module for Sloth.

:- use_module('Core', [tonf/2, tohnf/2]).

:- multifile     user_tohnf/2, user_tonf/2.
:- discontiguous user_tohnf/2, user_tonf/2.

%% --- Converts a character into its ASCII value.
%% ord :: Char -> Int
%% ord external
user_tohnf('Prelude_ord'(Char)#R, R) :-
	!,
	tonf(Char, NFChar),
	freeze(NFChar, (
			 char_code(NFChar, R)
		     )).

%% --- Converts an ASCII value into a character.
%% chr :: Int -> Char
%% chr external
user_tohnf('Prelude_chr'(Int)#R, R) :-
	!, 
	tonf(Int, NFInt),
	freeze(NFInt, (
			  char_code(R, Int)
		      )).

%% Integer operations.
user_tohnf('Prelude_+'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F + G),
		      freeze(B, true)
		  )).

user_tohnf('Prelude_-'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F - G),
		      freeze(B, true)
		  )).

user_tohnf('Prelude_*'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F * G),
		      freeze(B, true)
		  )).

user_tohnf('Prelude_div'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F / G),
		      freeze(B, true)
		  )).

user_tohnf((D mod E)#B,B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, B is F mod G),
		      freeze(B, true)
		  )
	      ).

%% Comparison operators.

user_tohnf('Prelude_<'(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, (
				    F < G -> 
				    B = 'Prelude_True' 
				; 
				    B = 'Prelude_False'
				)
			    )
		  )
	      ).

user_tohnf('Prelude_>'(D, E)#B, B) :- 
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, (
				    F > G -> 
				    B = 'Prelude_True' 
				; 
				    B = 'Prelude_False'
				)
			    )
		  )
	      ).

user_tohnf('Prelude_>='(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, (
				    F >= G -> 
				    B = 'Prelude_True' 
				; 
				    B = 'Prelude_False'
				)
			    )
		  )
	      ).

user_tohnf('Prelude_<='(D, E)#B, B) :-
	!,
	tonf(D,F),
	tonf(E,G),
	freeze(F, (
		      freeze(G, (
				    F =< G -> 
				    B = 'Prelude_True' 
				; 
				    B = 'Prelude_False'
				)
			    )
		  )
	      ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Igualdad
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

user_tohnf('Prelude_=:='(A, B)#'Prelude_Success', 'Prelude_Success') :-
	tohnf(A, A1),
	tohnf(B, B1),
	eq(A1, B1).

user_tohnf('Prelude_=='(A,B)#C,C) :- % Igualdad estricta (==)
	tohnf(A,A1),
	tohnf(B,B1),
	!,
	'=='(A1,B1,C).

user_tohnf('Prelude_success'#'Prelude_Success', 'Prelude_Success').

%% Concurrent constraints support.
user_tohnf('Prelude_&'(A, B)#C, C) :-
	%% Will be concurrent in the future.
	tohnf(A, R1),
	tohnf(B, R2),
	R1 = 'Prelude_Success',
	R2 = 'Prelude_Success',
	C = 'Prelude_Success'.

user_tohnf('Prelude_&>'(Constraint, Expr)#R, R) :-
	tohnf(Constraint, CR),
	CR = 'Prelude_Success',
	tonf(Expr, R).

user_tohnf('Prelude_tonf'(Expr)#R, R) :-
	tonf(Expr, R).

%% Nonvar primitive, it doens't work as now but it is a start.
user_tohnf('Prelude_nonvar'(A)#A, A) :-
	freeze(A, true).

'=='(A, B, 'Prelude_True') :-
	eq(A, B),
	!.

'=='(A, B, 'Prelude_False') :-
	\+ eq(A, B),
	!.


% eq(A,B):-(var(A);var(B)),!,A=B.
eq(A,B) :-
	var(A),
	!,
	tonf(B,A),
	!.

eq(A,B) :-
	var(B),
	!,
	tonf(A,B),
	!.

%% Fixme: Is this ok? -- egallego@babel
eq(A, A) :-
	A = pap(_, _, _)#_,
	!.

eq([],[]) :- 
	!.
eq([H|Ts], [A|Zs]) :- 
	!,
	tohnf(H,H1),
	tohnf(A,A1),
	eq(H1,A1),
	tohnf(Ts,Ts1),
	tohnf(Zs,Zs1),
	eq(Ts1,Zs1),
	!.

eq(A,B) :-
  A =.. [Ca | Alist],
  B =.. [Cb | Blist],
  !,
  Ca = Cb,
  eq(Alist, Blist),
  !.

eq(A, A) :-
	!.


eq_list([],[]).
eq_list([A|As], [B|Bs]) :-
	eq(A, B),
	!,
	eq_list(As, Bs),
	!.

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

