:- module(comp_manager, [compileFile/2, compileFilePrelude/2, rewrite/0]).

% Library modules:
:- use_module(library(dec10_io), [tell/1, told/0] ).
:- use_module(library(terms),    [atom_concat/2] ).

% Sloth modules:
:- use_module(symtab,     [clean_symtab/0] ).
:- use_module(scope,      [clean_scope/0] ).
:- use_module(cmdopts,    [cmd_options/2, set_option/2] ).
:- use_module(debug,      [dump_abstree/1, dump_deftrees/1, dump_rules/1,
	                   pretty_list_write/1, diff_lists/3]).

:- use_module(parser,     [parse/1] ).
:- use_module(unref,      [solve_all_refs/0, free_all_vars/0, 
	                   solve_all_type_refs/0, free_all_type_vars/0]).
:- use_module(lambda,     [desugar_lambdas/0] ).
:- use_module(localdefs,  [desugar_ldef/0] ).
:- use_module(verifier,   [verify_rules/0] ).
:- use_module(trans_lazy, [dump/0]).


% :- use_module(typer,      [check_infer_types/0] ).

% Compiler passes.

passes([solve_all_refs, desugar_ldef, desugar_lambdas, free_all_vars,
	solve_all_type_refs, free_all_type_vars]).

%% Main compilation function.
compileFile(In, Out) :-
	clean_symtab,
	clean_scope,
	parse(In),
	!,
	start_debug(Out),
	rewrite,
	stop_debug,
	!,
	tell(Out),
	dump,
	told.

compileFilePrelude(In, Out) :-
	set_option('-np', none),
	compileFile(In, Out).

rewrite :-
	passes(P),
	do_passes(P).

do_passes([]).
do_passes([P|PR]) :-
%	write(starting_pass(P)),
	call(P),
	do_debug(P),
	!,
	do_passes(PR).

:- dynamic last_data/1.

start_debug(File) :-
	cmd_options('-v', _),
	atom_concat([File, '.debug'], FileDebug),
	tell(FileDebug),
	dump_abstree(Abst),
	pretty_list_write(Abst),
	assert(last_data(Abst)),
	!.

start_debug(_) :- !.

stop_debug :-
	cmd_options('-v', _),
	retract(last_data(_)),
	dump_deftrees(Dt),
	pretty_list_write(Dt),
	dump_rules(Abst),
	pretty_list_write(Abst),
	told.

stop_debug :- !.

do_debug(Phase) :-
	(
	    cmd_options('-v', _) ->
	    write('\nPhase: '),
	    write(Phase),
	    write('\n'),
	    retract(last_data(OldAbs)),
	    dump_abstree(NewAbst),
	    assert(last_data(NewAbst)),
	    diff_lists(OldAbs, NewAbst, DAbs),
	    pretty_list_write(DAbs)
	;
	    true
	).

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

