:- module(main, [main/0]).

:- use_module(library(lists),   [append/3]).
:- use_module(library(system),  [file_exists/1]).
:- use_module(library(write),   [write/1]).
:- use_module(library(terms),   [atom_concat/2]).

:- use_module(error,        [do_error/1]     ).
:- use_module(comp_manager, [compileFile/2]  ).
:- use_module(cmdopts,      [process_options/0, cmd_options/2, help/0] ).
:- use_module(shell,        [start_shell/0]).

% egallego@babel: Sloth main module.
% Initial version: only compiles the file specified.

main :-
	catch(sloth_main, E, do_error(E)).

sloth_main :-
	process_options,
	(
	    cmd_options('File', File) ->
	    check_file(File),
	    make_pl_suffix(File, FileOut),
	    !,
	    compileFile(File, FileOut),
	    write('File '), write(File), 
	    write(' is compiled to '), write(FileOut), 
	    write('\n')
	;
	    cmd_options('-i', _),
	    start_shell
	).

sloth_main :-
	help.

% Checks the file exists.
check_file(File):-
	(
	    file_exists(File) -> 
	    true
	;    
	    atom_concat(['File does not exist: ', File], E),
	    throw(E)
        ).

% Changes a f.curry to f.pl
make_pl_suffix(OldFile, NewFile) :-
	atom_codes(OldFile, OFString),
	append(BaseName, ".curry", OFString),
	append(BaseName, ".pl", NFString),
	atom_codes(NewFile, NFString),!.

make_pl_suffix(_, _) :- !,
	throw('Error, filename must end in .curry').
	

%% Local Variables: 
%% mode: CIAO
%% update-version-comments: "off"
%% End:

